@@ -17,7 +17,8 @@
 
 #pragma once
 
-#include <Rcpp.h>
+#include <Rinternals.h>
+#include <Rmath.h>
 
 namespace RcppHoney {
 
@@ -36,4 +37,17 @@ struct na< double > {
     static const bool is_na(double val) {return R_IsNA(val);}
 };
 
+template< typename T >
+struct rtype {};
+
+template<>
+struct rtype< int > {
+    static const int value = INTSXP;
+};
+
+template<>
+struct rtype< double > {
+    static const int value = REALSXP;
+};
+
 } // namespace RcppHoney
